// Copyright (C) 2012-2013 Xilinx, Inc.  All rights reserved
// $Id: DiagramEmitter.h,v 1.2.2.9 2013/07/11 01:02:35 jcorbett Exp $

#ifndef Iv_DiagramEmitter_h
#define Iv_DiagramEmitter_h

#include "TUtilities/export/smart_ptr.hpp"
#include "Databases/Rf/export/Facade/Rf_Cell.h"
#include "Data.h"
#include "GroupSetFamily.h"
#include "Tile.h"
#include "ContourMap.h"
#include "DeviceModel.h"

typedef std::vector<Tile*> TileVec;
typedef TileVec::iterator TileVecIt;
typedef std::vector<size_t> IndexVec;
typedef IndexVec::iterator IndexVecIt;
typedef std::vector<double> CoordVec;
typedef std::map<Rf_Cell, CoordVec> BoundaryMap;
typedef BoundaryMap::iterator BoundaryMapIt;
typedef struct {
   double x;
   double y;
   std::string text;
   std::string style;
} Label;
typedef struct {
   bool operator()(const Label& a, const Label& b) {
      return a.text < b.text;
   }
} LabelLess;
typedef std::set<Label, LabelLess> LabelSet;
typedef LabelSet::iterator LabelSetIt;
typedef std::map<size_t, LabelSet> LabelMap;
typedef LabelMap::iterator LabelMapIt;

bool IsRectangular(const Rf_Cell site);

class DiagramEmitter {
public:
   static const double areaRangeOutlineWidth2;
   static const double areaRangeCornerRadius;
   static const double areaRangeStrokeOpacity;
   static const double areaRangeStrokeWidth;
   static const double boxHeight;
   static const double boxWidth;
   static const double busMacroStrokeOpacity;
   static const double tileStrokeOpacity;
   static const double unusedTileStrokeOpacity;
   static const double fillOpacity;
   static const double fontSize;
   static const double globalInset;
   static const double insetOpacity;
   static const double insetStrokeWidth;
   static const double lineHeightSmall;
   static const double lineHeightLarge;
   static const double progOpacity;
   static const double strokeWidth;
   static const double specialInset;
   static const double specialLeftMargin;
   static const double superSiteStrokeOpacity;
   static const double textBaselineOffset;
   static const double textLeftMargin;
   static const double unusedStrokeWidth;
   static const std::string globalStyle;
   static const std::string font;
   static const std::string logicStrokeColor;
   static const std::string unusedStyle;
   static const std::string superSiteStyle;
   static const std::string superSiteViolation;
   static const std::string shortStyle;
   static const std::string vertStyle;

public:
   DiagramEmitter(Data& data,
                  const boost::shared_ptr<ContourMap<Tile> >& contourMap,
                  const TileVec& grid,
                  GroupSetFamily& groupSetFamily,
                  const DeviceModel::CellSet& superIntSet,
                  const BoundaryMap& superBoundaryMap,
                  const BoundaryMap& superGlobalMap,
                  Data::GroupSet*& configCenterGroupSet);
   void SetHasConfigCenterViolation();
   bool IsDrawableSuperSite(const Rf_Cell superSite);
   bool IsShortableSuperSite(const Rf_Cell superSite);
   void GenerateDiagram();
   Tile* GetTile(Rf_Cell cell);
   double GetSmallZynqOffset();
private:
   void LabelGroupSets();
   const std::string GetGroupColor(Data::GroupSet* gs);
   const std::string GetGroupColorOpacity(Data::GroupSet* gs);
   const std::string EmitTiles();
   void AddLabel(double x, double y, std::string text, std::string style);
   const std::string EmitLabels();
   void DrawIOBSite(Rf_Cell site, std::string& boxes, std::string& labels);
   double GetSiteOffset(const std::string siteType) const;
   void DrawClockColumn(std::string& boxes);
   void DrawRectangularSuperSite(Rf_Cell site,
                                 std::string& boxes, std::string& labels);
   void DrawNonRectangularSuperSite(Rf_Cell site,
                                    std::string& boxes, std::string& labels);
   const std::string EmitSuperSites();
   const std::string AreaRangeRect(const std::string& comment,
                                   const std::string& groupColor,
                                   double xMin, double yMax,
                                   double xMax, double yMin);
   void EmitOneAreaRange(DeviceModel::SiteRect& rc,
                         const std::string& group,
                         const std::string& groupColor,
                         bool isFirst,
                         std::string& frames,
                         std::string& labels);
   const std::string EmitAreaRanges();
private:
   Data& data;
   const DeviceModel* deviceModel;
   const boost::shared_ptr<ContourMap<Tile> >& contourMap;
   const TileVec& grid;
   GroupSetFamily& groupSetFamily;
   IndexVec colorMap;
   const DeviceModel::CellSet& superIntSet;
   const BoundaryMap& superBoundaryMap;
   const BoundaryMap& superGlobalMap;
   LabelMap labelMap;
   Data::GroupSet*& configCenterGroupSet;
   bool hasConfigCenterViolation;
   double smallZynqOffset;
};

#endif
