// Copyright (C) 2006-2010 Xilinx, Inc.  All rights reserved
// $Id: PinGroup.h,v 1.11 2010/12/14 22:13:43 jcorbett Exp $
//
// R#check_package_pins, R#report_io_assignments, R#check_logic_tiles,
// R#check_interconnect_routing - Visitor for grid neighbors that
// collects pairs of neighbors from differing isolation groups along
// with enough information to give meaningful output.
//
// Note: it might be a slightly better design to split this class into
// two: one for package pins and another for IO buffers. It's slight
// because the only difference is that IO buffers carry one extra
// field.
#ifndef Iv_PinGroup_h
#define Iv_PinGroup_h

#include <string>
#include <vector>
#include <utility>

#include "main.h"

class PinGroup;
typedef std::pair<PinGroup*, PinGroup*> PinFault;

typedef std::vector<PinFault> PinFaultList;
typedef PinFaultList::iterator PinFaultIt;

// For package pins, the adjacent pins are defined as the immediate
// neighbors horizontally, vertically, and diagonally. IO buffers are
// arranged in vertical columns, so the adjacent neighbors are above
// and below. IO buffers have the additional restriction that no more
// than one isolation group can occupy a bank.
class PinGroup {
public:
   class PinFaultLess {
   public:
      bool operator()(const PinFault& a, const PinFault& b) const {
         if (a.first->pinName < b.first->pinName) return true;
         if (b.first->pinName < a.first->pinName) return false;
         if (a.second->pinName < b.second->pinName) return true;
         if (b.second->pinName < a.second->pinName) return false;
         return false;
      };
   };
public:
   PinGroup(std::string pin,
            std::string net,
            std::string group,
            PinFaultList& pinFaultList,
            int iobank = -1);
   void Visit(PinGroup* adjoined);
public:
   std::string pinName;
   std::string netName;
   std::string group;
   int iobank;                  // -1 if unused.

private:
   PinFaultList& faults;        // place to put the results
};

#endif
