// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: ArcCost.c,v 1.26.206.1 2013/05/10 00:04:05 jcorbett Exp $
#include "ArcCost.h"

#include <string>
#include <cstring>
#include <climits>

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.26.206.1 $";

// R#define_fence
Virtex4Cost::Virtex4Cost()
   : sourceFilter0(UCHAR_MAX, '\0'),
     sourceFilter1(UCHAR_MAX, '\0'),
     targetFilter0(UCHAR_MAX, '\0'),
     targetFilter1(UCHAR_MAX, '\0')
{
   CompileCostPatterns();
}

// R#define_fence - Compile the Virtex-4 arc cost regular expressions.
void Virtex4Cost::CompileCostPatterns() {
   const CostPat costArray[] = {
      // VCC and GND Costs
      {".+",                           "GND_WIRE",                        100},
      {"GND_WIRE",                     ".+",                              100},

      {".+",                           "VCC_WIRE",                        100},
      {"VCC_WIRE",                     ".+",                              100},

      // From BEST_LOGIC_OUTS Failure Analysis
      {"IMUX.*",                       "BEST_LOGIC_OUTS[0-7]",            3},

      // From ICLKP and OCLKP Failure Analysis
      {"BYP_INT_B1_INT",               "IOIS_ICLKP_0",                    1},
      {"BYP_INT_B3_INT",               "IOIS_ICLKP_1",                    1},
      {"BYP_INT_B5_INT",               "IOIS_OCLKP_0",                    1},
      {"BYP_INT_B6_INT",               "IOIS_OCLKP_1",                    1},

      {"IOIS_ICLKP_0",                 "BYP_INT_B1_INT",                  1},
      {"IOIS_ICLKP_1",                 "BYP_INT_B3_INT",                  1},
      {"IOIS_OCLKP_0",                 "BYP_INT_B5_INT",                  1},
      {"IOIS_OCLKP_1",                 "BYP_INT_B6_INT",                  1},

      {"IOIS_[IO]CLKP_[01]",           "IOIS_GCLKP[0-7]",                 1},
      {"IOIS_[IO]CLKP_[01]",           "IOIS_IOCLKP(_[NS])?[01]",         1},
      {"IOIS_[IO]CLKP_[01]",           "IOIS_RCLK_FORIO_P[01]",           1},

      {"IOIS_GCLKP[0-7]",              "IOIS_[IO]CLKP_[01]",              2},
      {"IOIS_IOCLKP(_[NS])?[01]",      "IOIS_[IO]CLKP_[01]",              2},
      {"IOIS_RCLK_FORIO_P[01]",        "IOIS_[IO]CLKP_[01]",              2},

      // From HCLK Failure Analysis
      // Note: [12]?[0-9]|3[01] matches 0, 1,  2, ..., 31.
      {"CLK_HROW_GCLK_BUFM([12]?[0-9]|3[01])", "CLK_HROW_HCLK_[LR]M[0-7]", 1},
      {"CLK_HROW_GCLK_BUFP([12]?[0-9]|3[01])", "CLK_HROW_HCLK_[LR]P[0-7]", 1},

      {"CLK_HROW_HCLK_[LR]M[0-7]", "CLK_HROW_GCLK_BUFM([12]?[0-9]|3[01])", 1},
      {"CLK_HROW_HCLK_[LR]P[0-7]", "CLK_HROW_GCLK_BUFP([12]?[0-9]|3[01])", 1},

      // From LEAF_GCLK Failure Analysis
      {"HCLK_G_HCLK[MP]0",             "HCLK_LEAF_GCLK0",                 1},
      {"HCLK_G_HCLK[MP]1",             "HCLK_LEAF_GCLK1",                 1},
      {"HCLK_G_HCLK[MP]2",             "HCLK_LEAF_GCLK2",                 1},
      {"HCLK_G_HCLK[MP]3",             "HCLK_LEAF_GCLK3",                 1},
      {"HCLK_G_HCLK[MP]4",             "HCLK_LEAF_GCLK4",                 1},
      {"HCLK_G_HCLK[MP]5",             "HCLK_LEAF_GCLK5",                 1},
      {"HCLK_G_HCLK[MP]6",             "HCLK_LEAF_GCLK6",                 1},
      {"HCLK_G_HCLK[MP]7",             "HCLK_LEAF_GCLK7",                 1},

      // From Reverse LEAF_GCLK Failure Analysis
      {"HCLK_LEAF_GCLK0",              "HCLK_G_HCLK[MP]0",                4},
      {"HCLK_LEAF_GCLK1",              "HCLK_G_HCLK[MP]1",                4},
      {"HCLK_LEAF_GCLK2",              "HCLK_G_HCLK[MP]2",                4},
      {"HCLK_LEAF_GCLK3",              "HCLK_G_HCLK[MP]3",                4},
      {"HCLK_LEAF_GCLK4",              "HCLK_G_HCLK[MP]4",                4},
      {"HCLK_LEAF_GCLK5",              "HCLK_G_HCLK[MP]5",                4},
      {"HCLK_LEAF_GCLK6",              "HCLK_G_HCLK[MP]6",                4},
      {"HCLK_LEAF_GCLK7",              "HCLK_G_HCLK[MP]7",                4},

      // From LEAF_RCLK Failure Analysis
      {"HCLK_RCLK0",                   "HCLK_LEAF_RCLK0",                 1},
      {"HCLK_RCLK1",                   "HCLK_LEAF_RCLK1",                 1},

      // From Viasat/Dave Schmolke design.
      // A NOR gate and three inverters would need to fail shorted to
      // connect these two nodes.
      {"HCLK_LEAF_RCLK[01]",           "HCLK_RCLK[01]",                   4},

      // From the MGT design
      {"BOUNCE[0-3]",                  "RCLK[01]",                        4},
      {"BOUNCE[0-3]",                  "GCLK[0-7]",                       4},

      // Pairs of lines below are forward direction / reverse direction

      // Global routing of BUFR and BUFIO clocks
      {"HCLK_IOIS_VRCLK(_[NS])?[01]",  "HCLK_IOIS_RCLK[01]",              1},
      {"HCLK_IOIS_RCLK[01]",           "HCLK_IOIS_VRCLK(_[NS])?[01]",     3},

      {"BYP_INT_B0",                   "BEST_LOGIC_OUTS[34]",             2},
      {"BEST_LOGIC_OUTS[34]",          "BYP_INT_B0",                      3},

      {"BYP_INT_B1",                   "BEST_LOGIC_OUTS[16]",             2},
      {"BEST_LOGIC_OUTS[16]",          "BYP_INT_B1",                      3},

      {"BYP_INT_B2",                   "BEST_LOGIC_OUTS[34]",             2},
      {"BEST_LOGIC_OUTS[34]",          "BYP_INT_B2",                      3},

      {"BYP_INT_B3",                   "BEST_LOGIC_OUTS[16]",             2},
      {"BEST_LOGIC_OUTS[16]",          "BYP_INT_B3",                      3},

      {"BYP_INT_B4",                   "BEST_LOGIC_OUTS[25]",             2},
      {"BEST_LOGIC_OUTS[25]",          "BYP_INT_B4",                      3},

      {"BYP_INT_B5",                   "BEST_LOGIC_OUTS[07]",             2},
      {"BEST_LOGIC_OUTS[07]",          "BYP_INT_B5",                      3},

      {"BYP_INT_B6",                   "BEST_LOGIC_OUTS[25]",             2},
      {"BEST_LOGIC_OUTS[25]",          "BYP_INT_B6",                      3},

      {"BYP_INT_B7",                   "BEST_LOGIC_OUTS[07]",             2},
      {"BEST_LOGIC_OUTS[07]",          "BYP_INT_B7",                      3},

      {".+",                           "CLK_B[0-3]_DCM0",                 2},

      // DCM Costs

      // Software Tile Conversion:
      //    HCLK_IOIS_LVDS_X42Y119 ==> cfg_hclk_iob_lvds_BD4.I0
      // Software/Hardware Conversion:
      // 1. HCLK_IOIS_LEAF_GCLK_P7 = [ leaf_gclk_p<7> ]
      // 2. HCLK_IOIS_G_HCLKP7 = [ g_hclkp<7> ]
      //

// Modified by SEM (8-June-2009).
// Added bit "0" to the next four sections.

      // Positive Terminal:
      {"HCLK_IOIS_LEAF_GCLK_P0",       "HCLK_IOIS_G_HCLKP0",              6},
      {"HCLK_IOIS_LEAF_GCLK_P1",       "HCLK_IOIS_G_HCLKP1",              6},
      {"HCLK_IOIS_LEAF_GCLK_P2",       "HCLK_IOIS_G_HCLKP2",              6},
      {"HCLK_IOIS_LEAF_GCLK_P3",       "HCLK_IOIS_G_HCLKP3",              6},
      {"HCLK_IOIS_LEAF_GCLK_P4",       "HCLK_IOIS_G_HCLKP4",              6},
      {"HCLK_IOIS_LEAF_GCLK_P5",       "HCLK_IOIS_G_HCLKP5",              6},
      {"HCLK_IOIS_LEAF_GCLK_P6",       "HCLK_IOIS_G_HCLKP6",              6},
      {"HCLK_IOIS_LEAF_GCLK_P7",       "HCLK_IOIS_G_HCLKP7",              6},

      // Minus Terminal:
      {"HCLK_IOIS_LEAF_GCLK_M0",       "HCLK_IOIS_G_HCLKP0",              7},
      {"HCLK_IOIS_LEAF_GCLK_M1",       "HCLK_IOIS_G_HCLKP1",              7},
      {"HCLK_IOIS_LEAF_GCLK_M2",       "HCLK_IOIS_G_HCLKP2",              7},
      {"HCLK_IOIS_LEAF_GCLK_M3",       "HCLK_IOIS_G_HCLKP3",              7},
      {"HCLK_IOIS_LEAF_GCLK_M4",       "HCLK_IOIS_G_HCLKP4",              7},
      {"HCLK_IOIS_LEAF_GCLK_M5",       "HCLK_IOIS_G_HCLKP5",              7},
      {"HCLK_IOIS_LEAF_GCLK_M6",       "HCLK_IOIS_G_HCLKP6",              7},
      {"HCLK_IOIS_LEAF_GCLK_M7",       "HCLK_IOIS_G_HCLKP7",              7},

      // Positive Terminal:
      {"HCLK_IOIS_G_HCLKP0",           "HCLK_IOIS_LEAF_GCLK_P0",          1},
      {"HCLK_IOIS_G_HCLKP1",           "HCLK_IOIS_LEAF_GCLK_P1",          1},
      {"HCLK_IOIS_G_HCLKP2",           "HCLK_IOIS_LEAF_GCLK_P2",          1},
      {"HCLK_IOIS_G_HCLKP3",           "HCLK_IOIS_LEAF_GCLK_P3",          1},
      {"HCLK_IOIS_G_HCLKP4",           "HCLK_IOIS_LEAF_GCLK_P4",          1},
      {"HCLK_IOIS_G_HCLKP5",           "HCLK_IOIS_LEAF_GCLK_P5",          1},
      {"HCLK_IOIS_G_HCLKP6",           "HCLK_IOIS_LEAF_GCLK_P6",          1},
      {"HCLK_IOIS_G_HCLKP7",           "HCLK_IOIS_LEAF_GCLK_P7",          1},

      // Minus Terminal:
      {"HCLK_IOIS_G_HCLKM0",           "HCLK_IOIS_LEAF_GCLK_P0",          1},
      {"HCLK_IOIS_G_HCLKM1",           "HCLK_IOIS_LEAF_GCLK_P1",          1},
      {"HCLK_IOIS_G_HCLKM2",           "HCLK_IOIS_LEAF_GCLK_P2",          1},
      {"HCLK_IOIS_G_HCLKM3",           "HCLK_IOIS_LEAF_GCLK_P3",          1},
      {"HCLK_IOIS_G_HCLKM4",           "HCLK_IOIS_LEAF_GCLK_P4",          1},
      {"HCLK_IOIS_G_HCLKM5",           "HCLK_IOIS_LEAF_GCLK_P5",          1},
      {"HCLK_IOIS_G_HCLKM6",           "HCLK_IOIS_LEAF_GCLK_P6",          1},
      {"HCLK_IOIS_G_HCLKM7",           "HCLK_IOIS_LEAF_GCLK_P7",          1},

      // The costs below in which source and target have matching wire
      // numbers from 0 through 31 were generated with the script
      // below.
      //
      // #!/perl -w
      // use strict;
      //
      // my @w = ('{"CLK_IOB_MUXED_CLKP_IN%", "CLK_IOB_MUXED_CLKP%", 2},',
      //          '{"CLK_IOB_MUXED_CLKP%", "CLK_IOB_MUXED_CLKP_IN%", 7},',
      //          '{"CLK_IOB_MUXED_CLKM_IN%", "CLK_IOB_MUXED_CLKM%", 2},',
      //          '{"CLK_IOB_MUXED_CLKM%", "CLK_IOB_MUXED_CLKM_IN%", 7},',
      //          '{"CLKV_DCM_MUXED_CLKP_IN%", "CLKV_DCM_MUXED_CLKP_OUT%", 1},',
      //          '{"CLKV_DCM_MUXED_CLKP_OUT%", "CLKV_DCM_MUXED_CLKP_IN%", 5},',
      //          '{"CLKV_DCM_MUXED_CLKM_IN%", "CLKV_DCM_MUXED_CLKM_OUT%", 1},',
      //          '{"CLKV_DCM_MUXED_CLKM_OUT%", "CLKV_DCM_MUXED_CLKM_IN%", 5},');
      //
      // foreach my $s (@w) {
      //     for (my $i = 0; $i < 32; ++$i) {
      //        my $o = $s;
      //        $o =~ s/%/$i/g;
      //        print "      $o\n";
      //     }
      //     print "\n";
      // }
      //
      // Software Tile Conversion:
      //    CLK_IOB_B_X21Y31 ==> clk_iob_bot_P35.Iclk_iob_top
      // Software/Hardware Conversion:
      // 1. CLK_IOB_MUXED_CLKP_IN24 = [ muxed_clkp_in<24> ]
      // 2. CLK_IOB_MUXED_CLKP24 = [ muxed_clkp<24>]
      // (n = 31:0 where 'n' must match on each side of the FROM/TO)

//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 4
      {"CLK_IOB_MUXED_CLKP_IN0",       "CLK_IOB_MUXED_CLKP0",             2},
      {"CLK_IOB_MUXED_CLKP_IN1",       "CLK_IOB_MUXED_CLKP1",             2},
      {"CLK_IOB_MUXED_CLKP_IN2",       "CLK_IOB_MUXED_CLKP2",             2},
      {"CLK_IOB_MUXED_CLKP_IN3",       "CLK_IOB_MUXED_CLKP3",             2},
      {"CLK_IOB_MUXED_CLKP_IN4",       "CLK_IOB_MUXED_CLKP4",             2},
      {"CLK_IOB_MUXED_CLKP_IN5",       "CLK_IOB_MUXED_CLKP5",             2},
      {"CLK_IOB_MUXED_CLKP_IN6",       "CLK_IOB_MUXED_CLKP6",             2},
      {"CLK_IOB_MUXED_CLKP_IN7",       "CLK_IOB_MUXED_CLKP7",             2},
      {"CLK_IOB_MUXED_CLKP_IN8",       "CLK_IOB_MUXED_CLKP8",             2},
      {"CLK_IOB_MUXED_CLKP_IN9",       "CLK_IOB_MUXED_CLKP9",             2},
      {"CLK_IOB_MUXED_CLKP_IN10",      "CLK_IOB_MUXED_CLKP10",            2},
      {"CLK_IOB_MUXED_CLKP_IN11",      "CLK_IOB_MUXED_CLKP11",            2},
      {"CLK_IOB_MUXED_CLKP_IN12",      "CLK_IOB_MUXED_CLKP12",            2},
      {"CLK_IOB_MUXED_CLKP_IN13",      "CLK_IOB_MUXED_CLKP13",            2},
      {"CLK_IOB_MUXED_CLKP_IN14",      "CLK_IOB_MUXED_CLKP14",            2},
      {"CLK_IOB_MUXED_CLKP_IN15",      "CLK_IOB_MUXED_CLKP15",            2},
      {"CLK_IOB_MUXED_CLKP_IN16",      "CLK_IOB_MUXED_CLKP16",            2},
      {"CLK_IOB_MUXED_CLKP_IN17",      "CLK_IOB_MUXED_CLKP17",            2},
      {"CLK_IOB_MUXED_CLKP_IN18",      "CLK_IOB_MUXED_CLKP18",            2},
      {"CLK_IOB_MUXED_CLKP_IN19",      "CLK_IOB_MUXED_CLKP19",            2},
      {"CLK_IOB_MUXED_CLKP_IN20",      "CLK_IOB_MUXED_CLKP20",            2},
      {"CLK_IOB_MUXED_CLKP_IN21",      "CLK_IOB_MUXED_CLKP21",            2},
      {"CLK_IOB_MUXED_CLKP_IN22",      "CLK_IOB_MUXED_CLKP22",            2},
      {"CLK_IOB_MUXED_CLKP_IN23",      "CLK_IOB_MUXED_CLKP23",            2},
      {"CLK_IOB_MUXED_CLKP_IN24",      "CLK_IOB_MUXED_CLKP24",            2},
      {"CLK_IOB_MUXED_CLKP_IN25",      "CLK_IOB_MUXED_CLKP25",            2},
      {"CLK_IOB_MUXED_CLKP_IN26",      "CLK_IOB_MUXED_CLKP26",            2},
      {"CLK_IOB_MUXED_CLKP_IN27",      "CLK_IOB_MUXED_CLKP27",            2},
      {"CLK_IOB_MUXED_CLKP_IN28",      "CLK_IOB_MUXED_CLKP28",            2},
      {"CLK_IOB_MUXED_CLKP_IN29",      "CLK_IOB_MUXED_CLKP29",            2},
      {"CLK_IOB_MUXED_CLKP_IN30",      "CLK_IOB_MUXED_CLKP30",            2},
      {"CLK_IOB_MUXED_CLKP_IN31",      "CLK_IOB_MUXED_CLKP31",            2},

//Modified by SEM(2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 4
      {"CLK_IOB_MUXED_CLKP0",          "CLK_IOB_MUXED_CLKP_IN0",          7},
      {"CLK_IOB_MUXED_CLKP1",          "CLK_IOB_MUXED_CLKP_IN1",          7},
      {"CLK_IOB_MUXED_CLKP2",          "CLK_IOB_MUXED_CLKP_IN2",          7},
      {"CLK_IOB_MUXED_CLKP3",          "CLK_IOB_MUXED_CLKP_IN3",          7},
      {"CLK_IOB_MUXED_CLKP4",          "CLK_IOB_MUXED_CLKP_IN4",          7},
      {"CLK_IOB_MUXED_CLKP5",          "CLK_IOB_MUXED_CLKP_IN5",          7},
      {"CLK_IOB_MUXED_CLKP6",          "CLK_IOB_MUXED_CLKP_IN6",          7},
      {"CLK_IOB_MUXED_CLKP7",          "CLK_IOB_MUXED_CLKP_IN7",          7},
      {"CLK_IOB_MUXED_CLKP8",          "CLK_IOB_MUXED_CLKP_IN8",          7},
      {"CLK_IOB_MUXED_CLKP9",          "CLK_IOB_MUXED_CLKP_IN9",          7},
      {"CLK_IOB_MUXED_CLKP10",         "CLK_IOB_MUXED_CLKP_IN10",         7},
      {"CLK_IOB_MUXED_CLKP11",         "CLK_IOB_MUXED_CLKP_IN11",         7},
      {"CLK_IOB_MUXED_CLKP12",         "CLK_IOB_MUXED_CLKP_IN12",         7},
      {"CLK_IOB_MUXED_CLKP13",         "CLK_IOB_MUXED_CLKP_IN13",         7},
      {"CLK_IOB_MUXED_CLKP14",         "CLK_IOB_MUXED_CLKP_IN14",         7},
      {"CLK_IOB_MUXED_CLKP15",         "CLK_IOB_MUXED_CLKP_IN15",         7},
      {"CLK_IOB_MUXED_CLKP16",         "CLK_IOB_MUXED_CLKP_IN16",         7},
      {"CLK_IOB_MUXED_CLKP17",         "CLK_IOB_MUXED_CLKP_IN17",         7},
      {"CLK_IOB_MUXED_CLKP18",         "CLK_IOB_MUXED_CLKP_IN18",         7},
      {"CLK_IOB_MUXED_CLKP19",         "CLK_IOB_MUXED_CLKP_IN19",         7},
      {"CLK_IOB_MUXED_CLKP20",         "CLK_IOB_MUXED_CLKP_IN20",         7},
      {"CLK_IOB_MUXED_CLKP21",         "CLK_IOB_MUXED_CLKP_IN21",         7},
      {"CLK_IOB_MUXED_CLKP22",         "CLK_IOB_MUXED_CLKP_IN22",         7},
      {"CLK_IOB_MUXED_CLKP23",         "CLK_IOB_MUXED_CLKP_IN23",         7},
      {"CLK_IOB_MUXED_CLKP24",         "CLK_IOB_MUXED_CLKP_IN24",         7},
      {"CLK_IOB_MUXED_CLKP25",         "CLK_IOB_MUXED_CLKP_IN25",         7},
      {"CLK_IOB_MUXED_CLKP26",         "CLK_IOB_MUXED_CLKP_IN26",         7},
      {"CLK_IOB_MUXED_CLKP27",         "CLK_IOB_MUXED_CLKP_IN27",         7},
      {"CLK_IOB_MUXED_CLKP28",         "CLK_IOB_MUXED_CLKP_IN28",         7},
      {"CLK_IOB_MUXED_CLKP29",         "CLK_IOB_MUXED_CLKP_IN29",         7},
      {"CLK_IOB_MUXED_CLKP30",         "CLK_IOB_MUXED_CLKP_IN30",         7},
      {"CLK_IOB_MUXED_CLKP31",         "CLK_IOB_MUXED_CLKP_IN31",         7},

//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 4
      {"CLK_IOB_MUXED_CLKM_IN0",       "CLK_IOB_MUXED_CLKM0",             2},
      {"CLK_IOB_MUXED_CLKM_IN1",       "CLK_IOB_MUXED_CLKM1",             2},
      {"CLK_IOB_MUXED_CLKM_IN2",       "CLK_IOB_MUXED_CLKM2",             2},
      {"CLK_IOB_MUXED_CLKM_IN3",       "CLK_IOB_MUXED_CLKM3",             2},
      {"CLK_IOB_MUXED_CLKM_IN4",       "CLK_IOB_MUXED_CLKM4",             2},
      {"CLK_IOB_MUXED_CLKM_IN5",       "CLK_IOB_MUXED_CLKM5",             2},
      {"CLK_IOB_MUXED_CLKM_IN6",       "CLK_IOB_MUXED_CLKM6",             2},
      {"CLK_IOB_MUXED_CLKM_IN7",       "CLK_IOB_MUXED_CLKM7",             2},
      {"CLK_IOB_MUXED_CLKM_IN8",       "CLK_IOB_MUXED_CLKM8",             2},
      {"CLK_IOB_MUXED_CLKM_IN9",       "CLK_IOB_MUXED_CLKM9",             2},
      {"CLK_IOB_MUXED_CLKM_IN10",      "CLK_IOB_MUXED_CLKM10",            2},
      {"CLK_IOB_MUXED_CLKM_IN11",      "CLK_IOB_MUXED_CLKM11",            2},
      {"CLK_IOB_MUXED_CLKM_IN12",      "CLK_IOB_MUXED_CLKM12",            2},
      {"CLK_IOB_MUXED_CLKM_IN13",      "CLK_IOB_MUXED_CLKM13",            2},
      {"CLK_IOB_MUXED_CLKM_IN14",      "CLK_IOB_MUXED_CLKM14",            2},
      {"CLK_IOB_MUXED_CLKM_IN15",      "CLK_IOB_MUXED_CLKM15",            2},
      {"CLK_IOB_MUXED_CLKM_IN16",      "CLK_IOB_MUXED_CLKM16",            2},
      {"CLK_IOB_MUXED_CLKM_IN17",      "CLK_IOB_MUXED_CLKM17",            2},
      {"CLK_IOB_MUXED_CLKM_IN18",      "CLK_IOB_MUXED_CLKM18",            2},
      {"CLK_IOB_MUXED_CLKM_IN19",      "CLK_IOB_MUXED_CLKM19",            2},
      {"CLK_IOB_MUXED_CLKM_IN20",      "CLK_IOB_MUXED_CLKM20",            2},
      {"CLK_IOB_MUXED_CLKM_IN21",      "CLK_IOB_MUXED_CLKM21",            2},
      {"CLK_IOB_MUXED_CLKM_IN22",      "CLK_IOB_MUXED_CLKM22",            2},
      {"CLK_IOB_MUXED_CLKM_IN23",      "CLK_IOB_MUXED_CLKM23",            2},
      {"CLK_IOB_MUXED_CLKM_IN24",      "CLK_IOB_MUXED_CLKM24",            2},
      {"CLK_IOB_MUXED_CLKM_IN25",      "CLK_IOB_MUXED_CLKM25",            2},
      {"CLK_IOB_MUXED_CLKM_IN26",      "CLK_IOB_MUXED_CLKM26",            2},
      {"CLK_IOB_MUXED_CLKM_IN27",      "CLK_IOB_MUXED_CLKM27",            2},
      {"CLK_IOB_MUXED_CLKM_IN28",      "CLK_IOB_MUXED_CLKM28",            2},
      {"CLK_IOB_MUXED_CLKM_IN29",      "CLK_IOB_MUXED_CLKM29",            2},
      {"CLK_IOB_MUXED_CLKM_IN30",      "CLK_IOB_MUXED_CLKM30",            2},
      {"CLK_IOB_MUXED_CLKM_IN31",      "CLK_IOB_MUXED_CLKM31",            2},

//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 4
      {"CLK_IOB_MUXED_CLKM0",          "CLK_IOB_MUXED_CLKM_IN0",          7},
      {"CLK_IOB_MUXED_CLKM1",          "CLK_IOB_MUXED_CLKM_IN1",          7},
      {"CLK_IOB_MUXED_CLKM2",          "CLK_IOB_MUXED_CLKM_IN2",          7},
      {"CLK_IOB_MUXED_CLKM3",          "CLK_IOB_MUXED_CLKM_IN3",          7},
      {"CLK_IOB_MUXED_CLKM4",          "CLK_IOB_MUXED_CLKM_IN4",          7},
      {"CLK_IOB_MUXED_CLKM5",          "CLK_IOB_MUXED_CLKM_IN5",          7},
      {"CLK_IOB_MUXED_CLKM6",          "CLK_IOB_MUXED_CLKM_IN6",          7},
      {"CLK_IOB_MUXED_CLKM7",          "CLK_IOB_MUXED_CLKM_IN7",          7},
      {"CLK_IOB_MUXED_CLKM8",          "CLK_IOB_MUXED_CLKM_IN8",          7},
      {"CLK_IOB_MUXED_CLKM9",          "CLK_IOB_MUXED_CLKM_IN9",          7},
      {"CLK_IOB_MUXED_CLKM10",         "CLK_IOB_MUXED_CLKM_IN10",         7},
      {"CLK_IOB_MUXED_CLKM11",         "CLK_IOB_MUXED_CLKM_IN11",         7},
      {"CLK_IOB_MUXED_CLKM12",         "CLK_IOB_MUXED_CLKM_IN12",         7},
      {"CLK_IOB_MUXED_CLKM13",         "CLK_IOB_MUXED_CLKM_IN13",         7},
      {"CLK_IOB_MUXED_CLKM14",         "CLK_IOB_MUXED_CLKM_IN14",         7},
      {"CLK_IOB_MUXED_CLKM15",         "CLK_IOB_MUXED_CLKM_IN15",         7},
      {"CLK_IOB_MUXED_CLKM16",         "CLK_IOB_MUXED_CLKM_IN16",         7},
      {"CLK_IOB_MUXED_CLKM17",         "CLK_IOB_MUXED_CLKM_IN17",         7},
      {"CLK_IOB_MUXED_CLKM18",         "CLK_IOB_MUXED_CLKM_IN18",         7},
      {"CLK_IOB_MUXED_CLKM19",         "CLK_IOB_MUXED_CLKM_IN19",         7},
      {"CLK_IOB_MUXED_CLKM20",         "CLK_IOB_MUXED_CLKM_IN20",         7},
      {"CLK_IOB_MUXED_CLKM21",         "CLK_IOB_MUXED_CLKM_IN21",         7},
      {"CLK_IOB_MUXED_CLKM22",         "CLK_IOB_MUXED_CLKM_IN22",         7},
      {"CLK_IOB_MUXED_CLKM23",         "CLK_IOB_MUXED_CLKM_IN23",         7},
      {"CLK_IOB_MUXED_CLKM24",         "CLK_IOB_MUXED_CLKM_IN24",         7},
      {"CLK_IOB_MUXED_CLKM25",         "CLK_IOB_MUXED_CLKM_IN25",         7},
      {"CLK_IOB_MUXED_CLKM26",         "CLK_IOB_MUXED_CLKM_IN26",         7},
      {"CLK_IOB_MUXED_CLKM27",         "CLK_IOB_MUXED_CLKM_IN27",         7},
      {"CLK_IOB_MUXED_CLKM28",         "CLK_IOB_MUXED_CLKM_IN28",         7},
      {"CLK_IOB_MUXED_CLKM29",         "CLK_IOB_MUXED_CLKM_IN29",         7},
      {"CLK_IOB_MUXED_CLKM30",         "CLK_IOB_MUXED_CLKM_IN30",         7},
      {"CLK_IOB_MUXED_CLKM31",         "CLK_IOB_MUXED_CLKM_IN31",         7},

      // CLK_DCM schematic identical at this level to CLK_IOB
      // (n = 31:0 where 'n' must match on each side of the FROM/TO)
      {"CLKV_DCM_MUXED_CLKP_IN0",      "CLKV_DCM_MUXED_CLKP_OUT0",        1},
      {"CLKV_DCM_MUXED_CLKP_IN1",      "CLKV_DCM_MUXED_CLKP_OUT1",        1},
      {"CLKV_DCM_MUXED_CLKP_IN2",      "CLKV_DCM_MUXED_CLKP_OUT2",        1},
      {"CLKV_DCM_MUXED_CLKP_IN3",      "CLKV_DCM_MUXED_CLKP_OUT3",        1},
      {"CLKV_DCM_MUXED_CLKP_IN4",      "CLKV_DCM_MUXED_CLKP_OUT4",        1},
      {"CLKV_DCM_MUXED_CLKP_IN5",      "CLKV_DCM_MUXED_CLKP_OUT5",        1},
      {"CLKV_DCM_MUXED_CLKP_IN6",      "CLKV_DCM_MUXED_CLKP_OUT6",        1},
      {"CLKV_DCM_MUXED_CLKP_IN7",      "CLKV_DCM_MUXED_CLKP_OUT7",        1},
      {"CLKV_DCM_MUXED_CLKP_IN8",      "CLKV_DCM_MUXED_CLKP_OUT8",        1},
      {"CLKV_DCM_MUXED_CLKP_IN9",      "CLKV_DCM_MUXED_CLKP_OUT9",        1},
      {"CLKV_DCM_MUXED_CLKP_IN10",     "CLKV_DCM_MUXED_CLKP_OUT10",       1},
      {"CLKV_DCM_MUXED_CLKP_IN11",     "CLKV_DCM_MUXED_CLKP_OUT11",       1},
      {"CLKV_DCM_MUXED_CLKP_IN12",     "CLKV_DCM_MUXED_CLKP_OUT12",       1},
      {"CLKV_DCM_MUXED_CLKP_IN13",     "CLKV_DCM_MUXED_CLKP_OUT13",       1},
      {"CLKV_DCM_MUXED_CLKP_IN14",     "CLKV_DCM_MUXED_CLKP_OUT14",       1},
      {"CLKV_DCM_MUXED_CLKP_IN15",     "CLKV_DCM_MUXED_CLKP_OUT15",       1},
      {"CLKV_DCM_MUXED_CLKP_IN16",     "CLKV_DCM_MUXED_CLKP_OUT16",       1},
      {"CLKV_DCM_MUXED_CLKP_IN17",     "CLKV_DCM_MUXED_CLKP_OUT17",       1},
      {"CLKV_DCM_MUXED_CLKP_IN18",     "CLKV_DCM_MUXED_CLKP_OUT18",       1},
      {"CLKV_DCM_MUXED_CLKP_IN19",     "CLKV_DCM_MUXED_CLKP_OUT19",       1},
      {"CLKV_DCM_MUXED_CLKP_IN20",     "CLKV_DCM_MUXED_CLKP_OUT20",       1},
      {"CLKV_DCM_MUXED_CLKP_IN21",     "CLKV_DCM_MUXED_CLKP_OUT21",       1},
      {"CLKV_DCM_MUXED_CLKP_IN22",     "CLKV_DCM_MUXED_CLKP_OUT22",       1},
      {"CLKV_DCM_MUXED_CLKP_IN23",     "CLKV_DCM_MUXED_CLKP_OUT23",       1},
      {"CLKV_DCM_MUXED_CLKP_IN24",     "CLKV_DCM_MUXED_CLKP_OUT24",       1},
      {"CLKV_DCM_MUXED_CLKP_IN25",     "CLKV_DCM_MUXED_CLKP_OUT25",       1},
      {"CLKV_DCM_MUXED_CLKP_IN26",     "CLKV_DCM_MUXED_CLKP_OUT26",       1},
      {"CLKV_DCM_MUXED_CLKP_IN27",     "CLKV_DCM_MUXED_CLKP_OUT27",       1},
      {"CLKV_DCM_MUXED_CLKP_IN28",     "CLKV_DCM_MUXED_CLKP_OUT28",       1},
      {"CLKV_DCM_MUXED_CLKP_IN29",     "CLKV_DCM_MUXED_CLKP_OUT29",       1},
      {"CLKV_DCM_MUXED_CLKP_IN30",     "CLKV_DCM_MUXED_CLKP_OUT30",       1},
      {"CLKV_DCM_MUXED_CLKP_IN31",     "CLKV_DCM_MUXED_CLKP_OUT31",       1},

      {"CLKV_DCM_MUXED_CLKP_OUT0",     "CLKV_DCM_MUXED_CLKP_IN0",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT1",     "CLKV_DCM_MUXED_CLKP_IN1",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT2",     "CLKV_DCM_MUXED_CLKP_IN2",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT3",     "CLKV_DCM_MUXED_CLKP_IN3",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT4",     "CLKV_DCM_MUXED_CLKP_IN4",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT5",     "CLKV_DCM_MUXED_CLKP_IN5",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT6",     "CLKV_DCM_MUXED_CLKP_IN6",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT7",     "CLKV_DCM_MUXED_CLKP_IN7",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT8",     "CLKV_DCM_MUXED_CLKP_IN8",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT9",     "CLKV_DCM_MUXED_CLKP_IN9",         5},
      {"CLKV_DCM_MUXED_CLKP_OUT10",    "CLKV_DCM_MUXED_CLKP_IN10",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT11",    "CLKV_DCM_MUXED_CLKP_IN11",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT12",    "CLKV_DCM_MUXED_CLKP_IN12",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT13",    "CLKV_DCM_MUXED_CLKP_IN13",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT14",    "CLKV_DCM_MUXED_CLKP_IN14",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT15",    "CLKV_DCM_MUXED_CLKP_IN15",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT16",    "CLKV_DCM_MUXED_CLKP_IN16",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT17",    "CLKV_DCM_MUXED_CLKP_IN17",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT18",    "CLKV_DCM_MUXED_CLKP_IN18",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT19",    "CLKV_DCM_MUXED_CLKP_IN19",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT20",    "CLKV_DCM_MUXED_CLKP_IN20",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT21",    "CLKV_DCM_MUXED_CLKP_IN21",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT22",    "CLKV_DCM_MUXED_CLKP_IN22",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT23",    "CLKV_DCM_MUXED_CLKP_IN23",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT24",    "CLKV_DCM_MUXED_CLKP_IN24",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT25",    "CLKV_DCM_MUXED_CLKP_IN25",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT26",    "CLKV_DCM_MUXED_CLKP_IN26",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT27",    "CLKV_DCM_MUXED_CLKP_IN27",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT28",    "CLKV_DCM_MUXED_CLKP_IN28",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT29",    "CLKV_DCM_MUXED_CLKP_IN29",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT30",    "CLKV_DCM_MUXED_CLKP_IN30",        5},
      {"CLKV_DCM_MUXED_CLKP_OUT31",    "CLKV_DCM_MUXED_CLKP_IN31",        5},

      {"CLKV_DCM_MUXED_CLKM_IN0",      "CLKV_DCM_MUXED_CLKM_OUT0",        1},
      {"CLKV_DCM_MUXED_CLKM_IN1",      "CLKV_DCM_MUXED_CLKM_OUT1",        1},
      {"CLKV_DCM_MUXED_CLKM_IN2",      "CLKV_DCM_MUXED_CLKM_OUT2",        1},
      {"CLKV_DCM_MUXED_CLKM_IN3",      "CLKV_DCM_MUXED_CLKM_OUT3",        1},
      {"CLKV_DCM_MUXED_CLKM_IN4",      "CLKV_DCM_MUXED_CLKM_OUT4",        1},
      {"CLKV_DCM_MUXED_CLKM_IN5",      "CLKV_DCM_MUXED_CLKM_OUT5",        1},
      {"CLKV_DCM_MUXED_CLKM_IN6",      "CLKV_DCM_MUXED_CLKM_OUT6",        1},
      {"CLKV_DCM_MUXED_CLKM_IN7",      "CLKV_DCM_MUXED_CLKM_OUT7",        1},
      {"CLKV_DCM_MUXED_CLKM_IN8",      "CLKV_DCM_MUXED_CLKM_OUT8",        1},
      {"CLKV_DCM_MUXED_CLKM_IN9",      "CLKV_DCM_MUXED_CLKM_OUT9",        1},
      {"CLKV_DCM_MUXED_CLKM_IN10",     "CLKV_DCM_MUXED_CLKM_OUT10",       1},
      {"CLKV_DCM_MUXED_CLKM_IN11",     "CLKV_DCM_MUXED_CLKM_OUT11",       1},
      {"CLKV_DCM_MUXED_CLKM_IN12",     "CLKV_DCM_MUXED_CLKM_OUT12",       1},
      {"CLKV_DCM_MUXED_CLKM_IN13",     "CLKV_DCM_MUXED_CLKM_OUT13",       1},
      {"CLKV_DCM_MUXED_CLKM_IN14",     "CLKV_DCM_MUXED_CLKM_OUT14",       1},
      {"CLKV_DCM_MUXED_CLKM_IN15",     "CLKV_DCM_MUXED_CLKM_OUT15",       1},
      {"CLKV_DCM_MUXED_CLKM_IN16",     "CLKV_DCM_MUXED_CLKM_OUT16",       1},
      {"CLKV_DCM_MUXED_CLKM_IN17",     "CLKV_DCM_MUXED_CLKM_OUT17",       1},
      {"CLKV_DCM_MUXED_CLKM_IN18",     "CLKV_DCM_MUXED_CLKM_OUT18",       1},
      {"CLKV_DCM_MUXED_CLKM_IN19",     "CLKV_DCM_MUXED_CLKM_OUT19",       1},
      {"CLKV_DCM_MUXED_CLKM_IN20",     "CLKV_DCM_MUXED_CLKM_OUT20",       1},
      {"CLKV_DCM_MUXED_CLKM_IN21",     "CLKV_DCM_MUXED_CLKM_OUT21",       1},
      {"CLKV_DCM_MUXED_CLKM_IN22",     "CLKV_DCM_MUXED_CLKM_OUT22",       1},
      {"CLKV_DCM_MUXED_CLKM_IN23",     "CLKV_DCM_MUXED_CLKM_OUT23",       1},
      {"CLKV_DCM_MUXED_CLKM_IN24",     "CLKV_DCM_MUXED_CLKM_OUT24",       1},
      {"CLKV_DCM_MUXED_CLKM_IN25",     "CLKV_DCM_MUXED_CLKM_OUT25",       1},
      {"CLKV_DCM_MUXED_CLKM_IN26",     "CLKV_DCM_MUXED_CLKM_OUT26",       1},
      {"CLKV_DCM_MUXED_CLKM_IN27",     "CLKV_DCM_MUXED_CLKM_OUT27",       1},
      {"CLKV_DCM_MUXED_CLKM_IN28",     "CLKV_DCM_MUXED_CLKM_OUT28",       1},
      {"CLKV_DCM_MUXED_CLKM_IN29",     "CLKV_DCM_MUXED_CLKM_OUT29",       1},
      {"CLKV_DCM_MUXED_CLKM_IN30",     "CLKV_DCM_MUXED_CLKM_OUT30",       1},
      {"CLKV_DCM_MUXED_CLKM_IN31",     "CLKV_DCM_MUXED_CLKM_OUT31",       1},

//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 5
      {"CLKV_DCM_MUXED_CLKM_OUT0",     "CLKV_DCM_MUXED_CLKM_IN0",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT1",     "CLKV_DCM_MUXED_CLKM_IN1",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT2",     "CLKV_DCM_MUXED_CLKM_IN2",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT3",     "CLKV_DCM_MUXED_CLKM_IN3",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT4",     "CLKV_DCM_MUXED_CLKM_IN4",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT5",     "CLKV_DCM_MUXED_CLKM_IN5",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT6",     "CLKV_DCM_MUXED_CLKM_IN6",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT7",     "CLKV_DCM_MUXED_CLKM_IN7",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT8",     "CLKV_DCM_MUXED_CLKM_IN8",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT9",     "CLKV_DCM_MUXED_CLKM_IN9",         5},
      {"CLKV_DCM_MUXED_CLKM_OUT10",    "CLKV_DCM_MUXED_CLKM_IN10",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT11",    "CLKV_DCM_MUXED_CLKM_IN11",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT12",    "CLKV_DCM_MUXED_CLKM_IN12",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT13",    "CLKV_DCM_MUXED_CLKM_IN13",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT14",    "CLKV_DCM_MUXED_CLKM_IN14",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT15",    "CLKV_DCM_MUXED_CLKM_IN15",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT16",    "CLKV_DCM_MUXED_CLKM_IN16",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT17",    "CLKV_DCM_MUXED_CLKM_IN17",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT18",    "CLKV_DCM_MUXED_CLKM_IN18",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT19",    "CLKV_DCM_MUXED_CLKM_IN19",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT20",    "CLKV_DCM_MUXED_CLKM_IN20",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT21",    "CLKV_DCM_MUXED_CLKM_IN21",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT22",    "CLKV_DCM_MUXED_CLKM_IN22",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT23",    "CLKV_DCM_MUXED_CLKM_IN23",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT24",    "CLKV_DCM_MUXED_CLKM_IN24",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT25",    "CLKV_DCM_MUXED_CLKM_IN25",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT26",    "CLKV_DCM_MUXED_CLKM_IN26",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT27",    "CLKV_DCM_MUXED_CLKM_IN27",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT28",    "CLKV_DCM_MUXED_CLKM_IN28",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT29",    "CLKV_DCM_MUXED_CLKM_IN29",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT30",    "CLKV_DCM_MUXED_CLKM_IN30",        5},
      {"CLKV_DCM_MUXED_CLKM_OUT31",    "CLKV_DCM_MUXED_CLKM_IN31",        5},

      // Software Tile Conversion:
      //    CLK_IOB_B_X21Y31 ==> clk_iob_bot_P35.Iclk_iob_top
      // Software/Hardware Conversion:
      // 1. CLK_IOB_MUXED_CLKP24 = [ muxed_clkp_in<24> ]
      // 2. CLK_IOB_IOB_BUFCLKP15 = [ iob_bufclkp<15> ]

      // (n = 31:0 with any combination on each side of the FROM/TO)


//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 4
//Note that the BUFCLK signals are 15:0 while the MUXED signals are 31:0 (still any combination)
//=================I only changed the cost, not the equation=================
      {"CLK_IOB_MUXED_CLKP([12]?[0-9]|3[01])",  "CLK_IOB_IOB_BUFCLKP([12]?[0-9]|3[01])",9},
      {"CLK_IOB_IOB_BUFCLKP([12]?[0-9]|3[01])", "CLK_IOB_MUXED_CLKP([12]?[0-9]|3[01])", 3},
      {"CLK_IOB_MUXED_CLKM([12]?[0-9]|3[01])",  "CLK_IOB_IOB_BUFCLKM([12]?[0-9]|3[01])",9},
      {"CLK_IOB_IOB_BUFCLKM([12]?[0-9]|3[01])", "CLK_IOB_MUXED_CLKM([12]?[0-9]|3[01])", 3},

      // CLK_DCM schematic identical at this level to CLK_IOB
      // CHANGE "IOB" to "DCM"in software name
      // CHANGE "IN" to "OUT" in software name
      // CHANGE "muxed_clkp" to "muxed_clkp_out" in schematic name
      // CHANGE "muxed_clkm" to "muxed_clkm_out" in schematic name

      // CLK_DCM schematic is identical at this level to CLK_IOB. The
      // only differences are in the names.  The software to hardware
      // mapping is:
      //    CLKV_DCM_MUXED_CLKP_OUT = [muxed_clkp_out]
      //    CLKV_DCM_DCM_OUTCLKP    = [dcm_outclkp]
      // (n = 31:0 with any combination on each side of the FROM/TO)


//Modified by SEM (2-June-2009).
//Reference "Updates_to_RPT059-1_4.doc" table on page 5
//Note that the MUXED signals are 31:0 while the OUTCLK signals are 23:0 (still any combination)
//=================I only changed the cost, not the equation=================
      {"CLKV_DCM_MUXED_CLKP_OUT([12]?[0-9]|3[01])", "CLKV_DCM_DCM_OUTCLKP([12]?[0-9]|3[01])",   3},
      {"CLKV_DCM_DCM_OUTCLKP([12]?[0-9]|3[01])",    "CLKV_DCM_MUXED_CLKP_OUT([12]?[0-9]|3[01])",2},
      {"CLKV_DCM_MUXED_CLKM_OUT([12]?[0-9]|3[01])", "CLKV_DCM_DCM_OUTCLKM([12]?[0-9]|3[01])",   3},
      {"CLKV_DCM_DCM_OUTCLKM([12]?[0-9]|3[01])",    "CLKV_DCM_MUXED_CLKM_OUT([12]?[0-9]|3[01])",2},


      // Costs related to CFG_HCLK_IOB_DCI

// Modified by SEM (8-June-2009).
// I replaced the entire section as I thought there were errors.
// I also added the costs for the IO clock that is from the local region
// North
      {"HCLK_IOIS_IOCLKP_N0",          "HCLK_IOIS_VIOCLKP_N0",            6},
      {"HCLK_IOIS_IOCLKM_N0",          "HCLK_IOIS_VIOCLKP_N0",            6},
      {"HCLK_IOIS_IOCLKP_N0",          "HCLK_IOIS_VIOCLKM_N0",            7},
      {"HCLK_IOIS_IOCLKM_N0",          "HCLK_IOIS_VIOCLKM_N0",            7},
      {"HCLK_IOIS_VIOCLK[MP]_N0",      "HCLK_IOIS_IOCLKP_N0",             1},
      {"HCLK_IOIS_VIOCLK[MP]_N0",      "HCLK_IOIS_IOCLKM_N0",             1},

      {"HCLK_IOIS_IOCLKP_N1",          "HCLK_IOIS_VIOCLKP_N1",            6},
      {"HCLK_IOIS_IOCLKM_N1",          "HCLK_IOIS_VIOCLKP_N1",            6},
      {"HCLK_IOIS_IOCLKP_N1",          "HCLK_IOIS_VIOCLKM_N1",            7},
      {"HCLK_IOIS_IOCLKM_N1",          "HCLK_IOIS_VIOCLKM_N1",            7},
      {"HCLK_IOIS_VIOCLK[MP]_N1",      "HCLK_IOIS_IOCLKP_N1",             1},
      {"HCLK_IOIS_VIOCLK[MP]_N1",      "HCLK_IOIS_IOCLKM_N1",             1},
// South
      {"HCLK_IOIS_IOCLKP_S0",          "HCLK_IOIS_VIOCLKP_S0",            6},

      {"HCLK_IOIS_IOCLKM_S0",          "HCLK_IOIS_VIOCLKP_S0",            6},
      {"HCLK_IOIS_IOCLKP_S0",          "HCLK_IOIS_VIOCLKM_S0",            7},
      {"HCLK_IOIS_IOCLKM_S0",          "HCLK_IOIS_VIOCLKM_S0",            7},
      {"HCLK_IOIS_VIOCLK[MP]_S0",      "HCLK_IOIS_IOCLKP_S0",             1},
      {"HCLK_IOIS_VIOCLK[MP]_S0",      "HCLK_IOIS_IOCLKM_S0",             1},

      {"HCLK_IOIS_IOCLKP_S1",          "HCLK_IOIS_VIOCLKP_S1",            6},
      {"HCLK_IOIS_IOCLKM_S1",          "HCLK_IOIS_VIOCLKP_S1",            6},
      {"HCLK_IOIS_IOCLKP_S1",          "HCLK_IOIS_VIOCLKM_S1",            7},
      {"HCLK_IOIS_IOCLKM_S1",          "HCLK_IOIS_VIOCLKM_S1",            7},
      {"HCLK_IOIS_VIOCLK[MP]_S1",      "HCLK_IOIS_IOCLKP_S1",             1},
      {"HCLK_IOIS_VIOCLK[MP]_S1",      "HCLK_IOIS_IOCLKM_S1",             1},
// Local
      {"HCLK_IOIS_IOCLKP0",            "HCLK_IOIS_VIOCLKP0",              6},
      {"HCLK_IOIS_IOCLKM0",            "HCLK_IOIS_VIOCLKP0",              6},
      {"HCLK_IOIS_IOCLKP0",            "HCLK_IOIS_VIOCLKM0",              7},
      {"HCLK_IOIS_IOCLKM0",            "HCLK_IOIS_VIOCLKM0",              7},
      {"HCLK_IOIS_VIOCLK[MP]0",        "HCLK_IOIS_IOCLKP0",               1},
      {"HCLK_IOIS_VIOCLK[MP]0",        "HCLK_IOIS_IOCLKM0",               1},

      {"HCLK_IOIS_IOCLKP1",            "HCLK_IOIS_VIOCLKP1",              6},
      {"HCLK_IOIS_IOCLKM1",            "HCLK_IOIS_VIOCLKP1",              6},
      {"HCLK_IOIS_IOCLKP1",            "HCLK_IOIS_VIOCLKM1",              7},
      {"HCLK_IOIS_IOCLKM1",            "HCLK_IOIS_VIOCLKM1",              7},
      {"HCLK_IOIS_VIOCLK[MP]1",        "HCLK_IOIS_IOCLKP1",               1},
      {"HCLK_IOIS_VIOCLK[MP]1",        "HCLK_IOIS_IOCLKM1",               1},
      // Generic Initial Cost
      // CLK_IOB_BUFCLKP<n> ==> CLK_IOB_IOB_CLKP<n>            COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_BUFCLKP0",             "CLK_IOB_IOB_CLKP0",               1},
      {"CLK_IOB_BUFCLKP1",             "CLK_IOB_IOB_CLKP1",               1},
      {"CLK_IOB_BUFCLKP2",             "CLK_IOB_IOB_CLKP2",               1},
      {"CLK_IOB_BUFCLKP3",             "CLK_IOB_IOB_CLKP3",               1},
      {"CLK_IOB_BUFCLKP4",             "CLK_IOB_IOB_CLKP4",               1},
      {"CLK_IOB_BUFCLKP5",             "CLK_IOB_IOB_CLKP5",               1},
      {"CLK_IOB_BUFCLKP6",             "CLK_IOB_IOB_CLKP6",               1},
      {"CLK_IOB_BUFCLKP7",             "CLK_IOB_IOB_CLKP7",               1},
      {"CLK_IOB_BUFCLKP8",             "CLK_IOB_IOB_CLKP8",               1},
      {"CLK_IOB_BUFCLKP9",             "CLK_IOB_IOB_CLKP9",               1},
      {"CLK_IOB_BUFCLKP10",            "CLK_IOB_IOB_CLKP10",              1},
      {"CLK_IOB_BUFCLKP11",            "CLK_IOB_IOB_CLKP11",              1},
      {"CLK_IOB_BUFCLKP12",            "CLK_IOB_IOB_CLKP12",              1},
      {"CLK_IOB_BUFCLKP13",            "CLK_IOB_IOB_CLKP13",              1},
      {"CLK_IOB_BUFCLKP14",            "CLK_IOB_IOB_CLKP14",              1},
      {"CLK_IOB_BUFCLKP15",            "CLK_IOB_IOB_CLKP15",              1},

      // Reverse Direction
      // CLK_IOB_IOB_CLKP<n> ==> CLK_IOB_BUFCLKP<n>            COST=5
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_CLKP0",           "CLK_IOB_BUFCLKP0",                 5},
      {"CLK_IOB_IOB_CLKP1",           "CLK_IOB_BUFCLKP1",                 5},
      {"CLK_IOB_IOB_CLKP2",           "CLK_IOB_BUFCLKP2",                 5},
      {"CLK_IOB_IOB_CLKP3",           "CLK_IOB_BUFCLKP3",                 5},
      {"CLK_IOB_IOB_CLKP4",           "CLK_IOB_BUFCLKP4",                 5},
      {"CLK_IOB_IOB_CLKP5",           "CLK_IOB_BUFCLKP5",                 5},
      {"CLK_IOB_IOB_CLKP6",           "CLK_IOB_BUFCLKP6",                 5},
      {"CLK_IOB_IOB_CLKP7",           "CLK_IOB_BUFCLKP7",                 5},
      {"CLK_IOB_IOB_CLKP8",           "CLK_IOB_BUFCLKP8",                 5},
      {"CLK_IOB_IOB_CLKP9",           "CLK_IOB_BUFCLKP9",                 5},
      {"CLK_IOB_IOB_CLKP10",          "CLK_IOB_BUFCLKP10",                5},
      {"CLK_IOB_IOB_CLKP11",          "CLK_IOB_BUFCLKP11",                5},
      {"CLK_IOB_IOB_CLKP12",          "CLK_IOB_BUFCLKP12",                5},
      {"CLK_IOB_IOB_CLKP13",          "CLK_IOB_BUFCLKP13",                5},
      {"CLK_IOB_IOB_CLKP14",          "CLK_IOB_BUFCLKP14",                5},
      {"CLK_IOB_IOB_CLKP15",          "CLK_IOB_BUFCLKP15",                5},

      // "M" side of the clock bus
      // CLK_IOB_BUFCLKM<n> ==> CLK_IOB_IOB_CLKM<n>            COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_BUFCLKM0",           "CLK_IOB_IOB_CLKM0",                 1},
      {"CLK_IOB_BUFCLKM1",           "CLK_IOB_IOB_CLKM1",                 1},
      {"CLK_IOB_BUFCLKM2",           "CLK_IOB_IOB_CLKM2",                 1},
      {"CLK_IOB_BUFCLKM3",           "CLK_IOB_IOB_CLKM3",                 1},
      {"CLK_IOB_BUFCLKM4",           "CLK_IOB_IOB_CLKM4",                 1},
      {"CLK_IOB_BUFCLKM5",           "CLK_IOB_IOB_CLKM5",                 1},
      {"CLK_IOB_BUFCLKM6",           "CLK_IOB_IOB_CLKM6",                 1},
      {"CLK_IOB_BUFCLKM7",           "CLK_IOB_IOB_CLKM7",                 1},
      {"CLK_IOB_BUFCLKM8",           "CLK_IOB_IOB_CLKM8",                 1},
      {"CLK_IOB_BUFCLKM9",           "CLK_IOB_IOB_CLKM9",                 1},
      {"CLK_IOB_BUFCLKM10",          "CLK_IOB_IOB_CLKM10",                1},
      {"CLK_IOB_BUFCLKM11",          "CLK_IOB_IOB_CLKM11",                1},
      {"CLK_IOB_BUFCLKM12",          "CLK_IOB_IOB_CLKM12",                1},
      {"CLK_IOB_BUFCLKM13",          "CLK_IOB_IOB_CLKM13",                1},
      {"CLK_IOB_BUFCLKM14",          "CLK_IOB_IOB_CLKM14",                1},
      {"CLK_IOB_BUFCLKM15",          "CLK_IOB_IOB_CLKM15",                1},

      // CLK_IOB_IOB_CLKM<n> ==> CLK_IOB_BUFCLKM<n>            COST=5
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_CLKM0",          "CLK_IOB_BUFCLKM0",                  5},
      {"CLK_IOB_IOB_CLKM1",          "CLK_IOB_BUFCLKM1",                  5},
      {"CLK_IOB_IOB_CLKM2",          "CLK_IOB_BUFCLKM2",                  5},
      {"CLK_IOB_IOB_CLKM3",          "CLK_IOB_BUFCLKM3",                  5},
      {"CLK_IOB_IOB_CLKM4",          "CLK_IOB_BUFCLKM4",                  5},
      {"CLK_IOB_IOB_CLKM5",          "CLK_IOB_BUFCLKM5",                  5},
      {"CLK_IOB_IOB_CLKM6",          "CLK_IOB_BUFCLKM6",                  5},
      {"CLK_IOB_IOB_CLKM7",          "CLK_IOB_BUFCLKM7",                  5},
      {"CLK_IOB_IOB_CLKM8",          "CLK_IOB_BUFCLKM8",                  5},
      {"CLK_IOB_IOB_CLKM9",          "CLK_IOB_BUFCLKM9",                  5},
      {"CLK_IOB_IOB_CLKM10",         "CLK_IOB_BUFCLKM10",                 5},
      {"CLK_IOB_IOB_CLKM11",         "CLK_IOB_BUFCLKM11",                 5},
      {"CLK_IOB_IOB_CLKM12",         "CLK_IOB_BUFCLKM12",                 5},
      {"CLK_IOB_IOB_CLKM13",         "CLK_IOB_BUFCLKM13",                 5},
      {"CLK_IOB_IOB_CLKM14",         "CLK_IOB_BUFCLKM14",                 5},
      {"CLK_IOB_IOB_CLKM15",         "CLK_IOB_BUFCLKM15",                 5},

      // HCLK_DCM_IOB_CLKP<n> ==> HCLK_DCM_IOB_CLKP_OUT<n>     COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"HCLK_DCM_IOB_CLKP0",         "HCLK_DCM_IOB_CLKP_OUT0",            1},
      {"HCLK_DCM_IOB_CLKP1",         "HCLK_DCM_IOB_CLKP_OUT1",            1},
      {"HCLK_DCM_IOB_CLKP2",         "HCLK_DCM_IOB_CLKP_OUT2",            1},
      {"HCLK_DCM_IOB_CLKP3",         "HCLK_DCM_IOB_CLKP_OUT3",            1},
      {"HCLK_DCM_IOB_CLKP4",         "HCLK_DCM_IOB_CLKP_OUT4",            1},
      {"HCLK_DCM_IOB_CLKP5",         "HCLK_DCM_IOB_CLKP_OUT5",            1},
      {"HCLK_DCM_IOB_CLKP6",         "HCLK_DCM_IOB_CLKP_OUT6",            1},
      {"HCLK_DCM_IOB_CLKP7",         "HCLK_DCM_IOB_CLKP_OUT7",            1},
      {"HCLK_DCM_IOB_CLKP8",         "HCLK_DCM_IOB_CLKP_OUT8",            1},
      {"HCLK_DCM_IOB_CLKP9",         "HCLK_DCM_IOB_CLKP_OUT9",            1},
      {"HCLK_DCM_IOB_CLKP10",        "HCLK_DCM_IOB_CLKP_OUT10",           1},
      {"HCLK_DCM_IOB_CLKP11",        "HCLK_DCM_IOB_CLKP_OUT11",           1},
      {"HCLK_DCM_IOB_CLKP12",        "HCLK_DCM_IOB_CLKP_OUT12",           1},
      {"HCLK_DCM_IOB_CLKP13",        "HCLK_DCM_IOB_CLKP_OUT13",           1},
      {"HCLK_DCM_IOB_CLKP14",        "HCLK_DCM_IOB_CLKP_OUT14",           1},
      {"HCLK_DCM_IOB_CLKP15",        "HCLK_DCM_IOB_CLKP_OUT15",           1},

      // Reverse Direction
      // HCLK_DCM_IOB_CLKP_OUT<n> ==> HCLK_DCM_IOB_CLKP<n>     COST=8
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"HCLK_DCM_IOB_CLKP_OUT0",     "HCLK_DCM_IOB_CLKP0",                8},
      {"HCLK_DCM_IOB_CLKP_OUT1",     "HCLK_DCM_IOB_CLKP1",                8},
      {"HCLK_DCM_IOB_CLKP_OUT2",     "HCLK_DCM_IOB_CLKP2",                8},
      {"HCLK_DCM_IOB_CLKP_OUT3",     "HCLK_DCM_IOB_CLKP3",                8},
      {"HCLK_DCM_IOB_CLKP_OUT4",     "HCLK_DCM_IOB_CLKP4",                8},
      {"HCLK_DCM_IOB_CLKP_OUT5",     "HCLK_DCM_IOB_CLKP5",                8},
      {"HCLK_DCM_IOB_CLKP_OUT6",     "HCLK_DCM_IOB_CLKP6",                8},
      {"HCLK_DCM_IOB_CLKP_OUT7",     "HCLK_DCM_IOB_CLKP7",                8},
      {"HCLK_DCM_IOB_CLKP_OUT8",     "HCLK_DCM_IOB_CLKP8",                8},
      {"HCLK_DCM_IOB_CLKP_OUT9",     "HCLK_DCM_IOB_CLKP9",                8},
      {"HCLK_DCM_IOB_CLKP_OUT10",    "HCLK_DCM_IOB_CLKP10",               8},
      {"HCLK_DCM_IOB_CLKP_OUT11",    "HCLK_DCM_IOB_CLKP11",               8},
      {"HCLK_DCM_IOB_CLKP_OUT12",    "HCLK_DCM_IOB_CLKP12",               8},
      {"HCLK_DCM_IOB_CLKP_OUT13",    "HCLK_DCM_IOB_CLKP13",               8},
      {"HCLK_DCM_IOB_CLKP_OUT14",    "HCLK_DCM_IOB_CLKP14",               8},
      {"HCLK_DCM_IOB_CLKP_OUT15",    "HCLK_DCM_IOB_CLKP15",               8},

      // "M" side of the clock bus
      // HCLK_DCM_IOB_CLKM<n> ==> HCLK_DCM_IOB_CLKM_OUT<n>     COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"HCLK_DCM_IOB_CLKM0",         "HCLK_DCM_IOB_CLKM_OUT0",            1},
      {"HCLK_DCM_IOB_CLKM1",         "HCLK_DCM_IOB_CLKM_OUT1",            1},
      {"HCLK_DCM_IOB_CLKM2",         "HCLK_DCM_IOB_CLKM_OUT2",            1},
      {"HCLK_DCM_IOB_CLKM3",         "HCLK_DCM_IOB_CLKM_OUT3",            1},
      {"HCLK_DCM_IOB_CLKM4",         "HCLK_DCM_IOB_CLKM_OUT4",            1},
      {"HCLK_DCM_IOB_CLKM5",         "HCLK_DCM_IOB_CLKM_OUT5",            1},
      {"HCLK_DCM_IOB_CLKM6",         "HCLK_DCM_IOB_CLKM_OUT6",            1},
      {"HCLK_DCM_IOB_CLKM7",         "HCLK_DCM_IOB_CLKM_OUT7",            1},
      {"HCLK_DCM_IOB_CLKM8",         "HCLK_DCM_IOB_CLKM_OUT8",            1},
      {"HCLK_DCM_IOB_CLKM9",         "HCLK_DCM_IOB_CLKM_OUT9",            1},
      {"HCLK_DCM_IOB_CLKM10",        "HCLK_DCM_IOB_CLKM_OUT10",           1},
      {"HCLK_DCM_IOB_CLKM11",        "HCLK_DCM_IOB_CLKM_OUT11",           1},
      {"HCLK_DCM_IOB_CLKM12",        "HCLK_DCM_IOB_CLKM_OUT12",           1},
      {"HCLK_DCM_IOB_CLKM13",        "HCLK_DCM_IOB_CLKM_OUT13",           1},
      {"HCLK_DCM_IOB_CLKM14",        "HCLK_DCM_IOB_CLKM_OUT14",           1},
      {"HCLK_DCM_IOB_CLKM15",        "HCLK_DCM_IOB_CLKM_OUT15",           1},

      // HCLK_DCM_IOB_CLKM_OUT<n> ==> HCLK_DCM_IOB_CLKM<n>     COST=8
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"HCLK_DCM_IOB_CLKM_OUT0",     "HCLK_DCM_IOB_CLKM0",                8},
      {"HCLK_DCM_IOB_CLKM_OUT1",     "HCLK_DCM_IOB_CLKM1",                8},
      {"HCLK_DCM_IOB_CLKM_OUT2",     "HCLK_DCM_IOB_CLKM2",                8},
      {"HCLK_DCM_IOB_CLKM_OUT3",     "HCLK_DCM_IOB_CLKM3",                8},
      {"HCLK_DCM_IOB_CLKM_OUT4",     "HCLK_DCM_IOB_CLKM4",                8},
      {"HCLK_DCM_IOB_CLKM_OUT5",     "HCLK_DCM_IOB_CLKM5",                8},
      {"HCLK_DCM_IOB_CLKM_OUT6",     "HCLK_DCM_IOB_CLKM6",                8},
      {"HCLK_DCM_IOB_CLKM_OUT7",     "HCLK_DCM_IOB_CLKM7",                8},
      {"HCLK_DCM_IOB_CLKM_OUT8",     "HCLK_DCM_IOB_CLKM8",                8},
      {"HCLK_DCM_IOB_CLKM_OUT9",     "HCLK_DCM_IOB_CLKM9",                8},
      {"HCLK_DCM_IOB_CLKM_OUT10",    "HCLK_DCM_IOB_CLKM10",               8},
      {"HCLK_DCM_IOB_CLKM_OUT11",    "HCLK_DCM_IOB_CLKM11",               8},
      {"HCLK_DCM_IOB_CLKM_OUT12",    "HCLK_DCM_IOB_CLKM12",               8},
      {"HCLK_DCM_IOB_CLKM_OUT13",    "HCLK_DCM_IOB_CLKM13",               8},
      {"HCLK_DCM_IOB_CLKM_OUT14",    "HCLK_DCM_IOB_CLKM14",               8},
      {"HCLK_DCM_IOB_CLKM_OUT15",    "HCLK_DCM_IOB_CLKM15",               8},

      // Generic Initial Cost
      // CLK_IOB_IOB_BUFCLKP<n> ==> CLK_IOB_IOB_CLKP<n>    COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_BUFCLKP0",       "CLK_IOB_IOB_CLKP0",                 1},
      {"CLK_IOB_IOB_BUFCLKP1",       "CLK_IOB_IOB_CLKP1",                 1},
      {"CLK_IOB_IOB_BUFCLKP2",       "CLK_IOB_IOB_CLKP2",                 1},
      {"CLK_IOB_IOB_BUFCLKP3",       "CLK_IOB_IOB_CLKP3",                 1},
      {"CLK_IOB_IOB_BUFCLKP4",       "CLK_IOB_IOB_CLKP4",                 1},
      {"CLK_IOB_IOB_BUFCLKP5",       "CLK_IOB_IOB_CLKP5",                 1},
      {"CLK_IOB_IOB_BUFCLKP6",       "CLK_IOB_IOB_CLKP6",                 1},
      {"CLK_IOB_IOB_BUFCLKP7",       "CLK_IOB_IOB_CLKP7",                 1},
      {"CLK_IOB_IOB_BUFCLKP8",       "CLK_IOB_IOB_CLKP8",                 1},
      {"CLK_IOB_IOB_BUFCLKP9",       "CLK_IOB_IOB_CLKP9",                 1},
      {"CLK_IOB_IOB_BUFCLKP10",      "CLK_IOB_IOB_CLKP10",                1},
      {"CLK_IOB_IOB_BUFCLKP11",      "CLK_IOB_IOB_CLKP11",                1},
      {"CLK_IOB_IOB_BUFCLKP12",      "CLK_IOB_IOB_CLKP12",                1},
      {"CLK_IOB_IOB_BUFCLKP13",      "CLK_IOB_IOB_CLKP13",                1},
      {"CLK_IOB_IOB_BUFCLKP14",      "CLK_IOB_IOB_CLKP14",                1},
      {"CLK_IOB_IOB_BUFCLKP15",      "CLK_IOB_IOB_CLKP15",                1},

      // Reverse Direction
      // CLK_IOB_IOB_CLKP<n> ==> CLK_IOB_IOB_BUFCLKP<n>    COST=5
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_CLKP0",          "CLK_IOB_IOB_BUFCLKP0",              5},
      {"CLK_IOB_IOB_CLKP1",          "CLK_IOB_IOB_BUFCLKP1",              5},
      {"CLK_IOB_IOB_CLKP2",          "CLK_IOB_IOB_BUFCLKP2",              5},
      {"CLK_IOB_IOB_CLKP3",          "CLK_IOB_IOB_BUFCLKP3",              5},
      {"CLK_IOB_IOB_CLKP4",          "CLK_IOB_IOB_BUFCLKP4",              5},
      {"CLK_IOB_IOB_CLKP5",          "CLK_IOB_IOB_BUFCLKP5",              5},
      {"CLK_IOB_IOB_CLKP6",          "CLK_IOB_IOB_BUFCLKP6",              5},
      {"CLK_IOB_IOB_CLKP7",          "CLK_IOB_IOB_BUFCLKP7",              5},
      {"CLK_IOB_IOB_CLKP8",          "CLK_IOB_IOB_BUFCLKP8",              5},
      {"CLK_IOB_IOB_CLKP9",          "CLK_IOB_IOB_BUFCLKP9",              5},
      {"CLK_IOB_IOB_CLKP10",         "CLK_IOB_IOB_BUFCLKP10",             5},
      {"CLK_IOB_IOB_CLKP11",         "CLK_IOB_IOB_BUFCLKP11",             5},
      {"CLK_IOB_IOB_CLKP12",         "CLK_IOB_IOB_BUFCLKP12",             5},
      {"CLK_IOB_IOB_CLKP13",         "CLK_IOB_IOB_BUFCLKP13",             5},
      {"CLK_IOB_IOB_CLKP14",         "CLK_IOB_IOB_BUFCLKP14",             5},
      {"CLK_IOB_IOB_CLKP15",         "CLK_IOB_IOB_BUFCLKP15",             5},

      // "M" side of the clock bus
      // CLK_IOB_IOB_BUFCLKM<n> ==> CLK_IOB_IOB_CLKM<n>   COST=1
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_BUFCLKM0",       "CLK_IOB_IOB_CLKM0",                 1},
      {"CLK_IOB_IOB_BUFCLKM1",       "CLK_IOB_IOB_CLKM1",                 1},
      {"CLK_IOB_IOB_BUFCLKM2",       "CLK_IOB_IOB_CLKM2",                 1},
      {"CLK_IOB_IOB_BUFCLKM3",       "CLK_IOB_IOB_CLKM3",                 1},
      {"CLK_IOB_IOB_BUFCLKM4",       "CLK_IOB_IOB_CLKM4",                 1},
      {"CLK_IOB_IOB_BUFCLKM5",       "CLK_IOB_IOB_CLKM5",                 1},
      {"CLK_IOB_IOB_BUFCLKM6",       "CLK_IOB_IOB_CLKM6",                 1},
      {"CLK_IOB_IOB_BUFCLKM7",       "CLK_IOB_IOB_CLKM7",                 1},
      {"CLK_IOB_IOB_BUFCLKM8",       "CLK_IOB_IOB_CLKM8",                 1},
      {"CLK_IOB_IOB_BUFCLKM9",       "CLK_IOB_IOB_CLKM9",                 1},
      {"CLK_IOB_IOB_BUFCLKM10",      "CLK_IOB_IOB_CLKM10",                1},
      {"CLK_IOB_IOB_BUFCLKM11",      "CLK_IOB_IOB_CLKM11",                1},
      {"CLK_IOB_IOB_BUFCLKM12",      "CLK_IOB_IOB_CLKM12",                1},
      {"CLK_IOB_IOB_BUFCLKM13",      "CLK_IOB_IOB_CLKM13",                1},
      {"CLK_IOB_IOB_BUFCLKM14",      "CLK_IOB_IOB_CLKM14",                1},
      {"CLK_IOB_IOB_BUFCLKM15",      "CLK_IOB_IOB_CLKM15",                1},

      // CLK_IOB_IOB_CLKM<n> ==> CLK_IOB_IOB_BUFCLKM<n>   COST=5
      // n = <15 downto 0>
      // n must match on both sides of cost
      {"CLK_IOB_IOB_CLKM0",          "CLK_IOB_IOB_BUFCLKM0",              5},
      {"CLK_IOB_IOB_CLKM1",          "CLK_IOB_IOB_BUFCLKM1",              5},
      {"CLK_IOB_IOB_CLKM2",          "CLK_IOB_IOB_BUFCLKM2",              5},
      {"CLK_IOB_IOB_CLKM3",          "CLK_IOB_IOB_BUFCLKM3",              5},
      {"CLK_IOB_IOB_CLKM4",          "CLK_IOB_IOB_BUFCLKM4",              5},
      {"CLK_IOB_IOB_CLKM5",          "CLK_IOB_IOB_BUFCLKM5",              5},
      {"CLK_IOB_IOB_CLKM6",          "CLK_IOB_IOB_BUFCLKM6",              5},
      {"CLK_IOB_IOB_CLKM7",          "CLK_IOB_IOB_BUFCLKM7",              5},
      {"CLK_IOB_IOB_CLKM8",          "CLK_IOB_IOB_BUFCLKM8",              5},
      {"CLK_IOB_IOB_CLKM9",          "CLK_IOB_IOB_BUFCLKM9",              5},
      {"CLK_IOB_IOB_CLKM10",         "CLK_IOB_IOB_BUFCLKM10",             5},
      {"CLK_IOB_IOB_CLKM11",         "CLK_IOB_IOB_BUFCLKM11",             5},
      {"CLK_IOB_IOB_CLKM12",         "CLK_IOB_IOB_BUFCLKM12",             5},
      {"CLK_IOB_IOB_CLKM13",         "CLK_IOB_IOB_BUFCLKM13",             5},
      {"CLK_IOB_IOB_CLKM14",         "CLK_IOB_IOB_BUFCLKM14",             5},
      {"CLK_IOB_IOB_CLKM15",         "CLK_IOB_IOB_BUFCLKM15",             5},

// Modified by SEM (3-May-2010).
      // Virtex-4 Cost Analysis GDLS / XC4VFX100-10FF1152
      {"SYS_MON_GIOB0",              "MONITOR_CONVST",                    2},
      // '[1-9]|1[0-5]' matches any of 1, 2, 3, ..., 15.
      {"SYS_MON_GIOB([1-9]|1[0-5])", "MONITOR_CONVST",                    3},
      {"MONITOR_CONVST",             "SYS_MON_GIOB0",                     10},
      {"MONITOR_CONVST",             "SYS_MON_GIOB([1-9]|1[0-5])",        11},

// Modified by SEM (10-May-2010).
      {"HCLK_DCM_IOB_CLKP0",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT0",      1},
      {"HCLK_DCM_IOB_CLKP1",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT1",      1},
      {"HCLK_DCM_IOB_CLKP2",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT2",      1},
      {"HCLK_DCM_IOB_CLKP3",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT3",      1},
      {"HCLK_DCM_IOB_CLKP4",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT4",      1},
      {"HCLK_DCM_IOB_CLKP5",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT5",      1},
      {"HCLK_DCM_IOB_CLKP6",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT6",      1},
      {"HCLK_DCM_IOB_CLKP7",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT7",      1},
      {"HCLK_DCM_IOB_CLKP8",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT8",      1},
      {"HCLK_DCM_IOB_CLKP9",         "HCLK_DCM_IOB_CLKP(_UP)?_OUT9",      1},
      {"HCLK_DCM_IOB_CLKP10",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT10",     1},
      {"HCLK_DCM_IOB_CLKP11",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT11",     1},
      {"HCLK_DCM_IOB_CLKP12",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT12",     1},
      {"HCLK_DCM_IOB_CLKP13",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT13",     1},
      {"HCLK_DCM_IOB_CLKP14",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT14",     1},
      {"HCLK_DCM_IOB_CLKP15",        "HCLK_DCM_IOB_CLKP(_UP)?_OUT15",     1},

      {"HCLK_DCM_IOB_CLKM0",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT0",      1},
      {"HCLK_DCM_IOB_CLKM1",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT1",      1},
      {"HCLK_DCM_IOB_CLKM2",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT2",      1},
      {"HCLK_DCM_IOB_CLKM3",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT3",      1},
      {"HCLK_DCM_IOB_CLKM4",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT4",      1},
      {"HCLK_DCM_IOB_CLKM5",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT5",      1},
      {"HCLK_DCM_IOB_CLKM6",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT6",      1},
      {"HCLK_DCM_IOB_CLKM7",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT7",      1},
      {"HCLK_DCM_IOB_CLKM8",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT8",      1},
      {"HCLK_DCM_IOB_CLKM9",         "HCLK_DCM_IOB_CLKM(_UP)?_OUT9",      1},
      {"HCLK_DCM_IOB_CLKM10",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT10",     1},
      {"HCLK_DCM_IOB_CLKM11",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT11",     1},
      {"HCLK_DCM_IOB_CLKM12",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT12",     1},
      {"HCLK_DCM_IOB_CLKM13",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT13",     1},
      {"HCLK_DCM_IOB_CLKM14",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT14",     1},
      {"HCLK_DCM_IOB_CLKM15",        "HCLK_DCM_IOB_CLKM(_UP)?_OUT15",     1},

      {"DCM_GIOB([0-9]|1[0-5])",     "DCM_ADV_CLK(IN|FB)",                2}
   };
   const long costPatCount(sizeof(costArray) / sizeof(costArray[0]));
   // Initialize Boolean vectors based on the first two characters of the source
   // and target patterns to spare us from performing an expensive
   // batch of regular expression tests that are guaranteed to
   // fail. This optimization reduces the run-time by approximately 50%.
   for (long i(0); i < costPatCount; ++i) {
      sourceFilter0.at((size_t) costArray[i].sourcePat[0]) = true;
      sourceFilter1.at((size_t) costArray[i].sourcePat[1]) = true;
      targetFilter0.at((size_t) costArray[i].targetPat[0]) = true;
      targetFilter1.at((size_t) costArray[i].targetPat[1]) = true;
   }
   for (long i(0); i < costPatCount; ++i) {
      try {
         // Both constArray[i].sourcePat and constArray[i].targetPat
         // have type const char*, however the Visual C++ 2003
         // compiler at version 7.10.3077 appears to need to be
         // reminded of this. This issue does not appear in Visual C++
         // 2005.
         WireNamePat source(string((const char*) costArray[i].sourcePat));
         WireNamePat target(string((const char*) costArray[i].targetPat));
         Port::UINT32 cost(costArray[i].cost);
         ArcCostListIt aclIt(find_if(arcCostList.begin(),
                                     arcCostList.end(),
                                     ArcCostMatchSource(source)));
         if (aclIt == arcCostList.end()) {
            arcCostList
               .push_back(make_pair(source,
                                    WireCostList(1,
                                                 make_pair(target, cost))));
         } else {
            aclIt->second.push_back(make_pair(target, cost));
         }
      } catch (boost::bad_pattern e) {
         FatalInternal(10, "Invalid regular expression syntax at i="
                       + ToString(i) + ":\n" + e.what());
      }
   }
}

// R#define_fence - Initially we thought that it would be sufficient
// to assign costs to wires based on their type codes.
//
// Without the additional information supplied below, the search would
// find long failing paths of zero nominal cost after a very long
// search. The search found a loop hole in our prior cost
// function. Since the software model is silent on the number of
// configuration memory cell (CMC) changes required to accomplish
// things, the work here is based on analysis of the Virtex4
// schematics. The analysis showed that for all the wires named below
// at least two CMC changes would be required to get information onto
// the wire.
//
// This function returns a lower bound the cost associated with a
// wire. These wires were the easiest ones to analyze and assign costs
// to. It is likely that many other wires have greater cost than this
// function returns, however we did not need to assign greater costs
// and therefore have not done the analysis.
//
// Specifically, it would take at least two faults to get information
// on to any of the the following 44 wires:
//   BYP_INT_B0, BYP_INT_B1, BYP_INT_B2, ..., BYP_INT_B7,
//   BOUNCE0, BOUNCE1, BOUNCE2 and BOUNCE3,
//   IMUX_B0, IMUX_B1, IMUX_B2, ..., IMUX_B31.
//
// This function takes an arc, a string that may be a prefix of the
// associated wire name, and the width of the bus corresponding to a
// matching wire name, and a suffix. It fetches the name of the wire,
// compares the prefix and the bus pin number. If both match, it
// returns a cost of one, otherwise it returns a cost of zero.
//
// Although the arc records the direction of current flow IVT is
// considering, the cost for using the arc in the reverse of the usual
// direction is no less than the cost of using it in the normal
// direction and sometimes greater, so the cost is calculated for the
// normal current flow direction. In order to go past a driver in the
// wrong direction, the driver transistor would have to fail, however
// it is not necessary to model that cost.
int Virtex4Cost::MatchBusWire(RevArc& arc,
                              const char* prefix,
                              Port::UINT32 prefixLen,
                              long busMax,
                              const char* suffix,
                              Port::UINT32 suffixLen)
{
   Port_String tile, wire;
   Rf_Node node(arc.GetTarget());
   Port::UINT32 wireIndex(arc.GetTargetIndex());
   node.GetWireNameAndTileInfo(wireIndex, tile, wire);

   // Since this is called in the innermost loop, it needs to be fast,
   // so I'm not converting from Port::StringBase and not using
   // regular expressions.

   // Shortest match will have prefix, a single digit, and suffix.
   Port::UINT32 lenMin(prefixLen + 1U + suffixLen);

   // Longest match will have prefix, two digits, and suffix.
   Port::UINT32
      lenMax(prefixLen + (9U < Port::UINT32(busMax) ? 2U : 1U) + suffixLen);

   if (wire.GetLength() < lenMin || lenMax < wire.GetLength()) return 0;

   // Convert the left-most character of the bus number to a long integer.
   long busPin(long(wire.GetChar(prefixLen)) - long('0'));

   // If the test below fails, the character wasn't a digit.
   if (busPin < 0 || 9 < busPin) return 0;

   // If busMax is 9 or less, then lenMin will equal lenMax and we only
   // have one digit to convert. Otherwise, get the next digit.
   if (lenMin < lenMax && wire.GetLength() == lenMax) {
      long unitsPlace(long(wire.GetChar(prefixLen + 1U)) - long('0'));
      busPin = 10 * busPin + unitsPlace;
      if (busPin < 0 || busMax < busPin) return 0;
   }
   if (strncmp(wire, prefix, prefixLen) != 0) return 0;
   Port::UINT32 suffixOffset(wire.GetLength() - suffixLen);
   if (0 < suffixLen
       && strncmp(suffixOffset + (const char*) wire, suffix, suffixLen) != 0)
   {
      return 0;
   }
   return 1;
}

ArcCost::~ArcCost() {
}

// R#define_fence - The bit count is the cost of getting off the first
// node of the arc and onto the second.
int Virtex4Cost::BitCount(RevArc& arc) {
   const string& source(arc.GetSourceName());
   const string& target(arc.GetTargetName());

   // Only perform regular expression tests if the first two
   // characters of the source or target could possibly match a
   // pattern.
   if ((sourceFilter0[source[0]] && sourceFilter1[source[1]])
       || (targetFilter0[target[0]] && targetFilter1[target[1]]))
   {
      for (ArcCostList::size_type i(0); i < arcCostList.size(); ++i) {
         if (boost::regex_match(source, arcCostList[i].first)) {
            for (WireCostListIt wcIt(arcCostList[i].second.begin());
                 wcIt != arcCostList[i].second.end();
                 ++wcIt)
            {
               if (boost::regex_match(target, wcIt->first)) {
                  return wcIt->second;
               }
            }
         }
      }
   }
   int offCount(0);
   Port::UINT32 sourceCode(arc.GetSource().GetTypeCode());
   Port::UINT32 targetCode(arc.GetTarget().GetTypeCode());
   switch (sourceCode) {
   case Virtex4Cost::OMUX:
   case Virtex4Cost::DOUBLE:
   case Virtex4Cost::HUNIHEX:
   case Virtex4Cost::VUNIHEX:
      // Getting off of an OMUX, DOUBLE, or HEX and onto a
      // BEST_LOGIC_OUTS(7:0) would require going backwards through a
      // driver.
      // If the destination is anything else, the cost is 1.
      if (MatchBusWire(arc,
                       "BEST_LOGIC_OUTS",
                       sizeof("BEST_LOGIC_OUTS")/sizeof(char) - 1,
                       7))
      {
         return 4;
      } else {
         offCount = 1;
      }
      break;
   case Virtex4Cost::HLONG:
   case Virtex4Cost::VLONG:
      offCount = 1;
      break;
   default:
      offCount = 0;
   };
   int onCount(0);
   switch (targetCode) {
   case Virtex4Cost::OMUX:
   case Virtex4Cost::DOUBLE:
   case Virtex4Cost::HUNIHEX:
   case Virtex4Cost::VUNIHEX:
   case Virtex4Cost::HLONG:
   case Virtex4Cost::VLONG:
      onCount = 2;
      break;
   case Virtex4Cost::PINBOUNCE:
   {
      // BYP_INT_B0, BYP_INT_B1, BYP_INT_B2, ..., BYP_INT_B7 and also
      // BYP_INT_B0_INT, BYP_INT_B1_INT, BYP_INT_B2_INT, ...,
      // BYP_INT_B7_INT have cost 2 in the forward direction and 1 in
      // the reverse direction.

      // The '_INT' suffix appear to be a name change as the node passes
      // from tile to tile.

      onCount =
         MatchBusWire(arc,
                      "BYP_INT_B",
                      sizeof("BYP_INT_B")/sizeof(char) - 1,
                      7)
         + MatchBusWire(arc,
                        "BYP_INT_B",
                        sizeof("BYP_INT_B")/sizeof(char) - 1,
                        7,
                        "_INT",
                        sizeof("_INT")/sizeof(char) - 1);
      if (arc.GetDirection() == RevArc::Forward) {
         onCount *= 2;
      } else if (arc.GetDirection() == RevArc::Reverse) {
         onCount *= 1;
      } else {
         // RevArc instances are immutable. Both constructors set
         // the direction field to a valid value. If we ever get here
         // we've got memory corruption.
         FatalInternal(82, "Invalid RevArc direction.");
      }
      break;
   }
   case Virtex4Cost::BOUNCEIN:
   {
      // BOUNCE0, BOUNCE1, BOUNCE2 and BOUNCE3 have cost 2.
      onCount = 2 * MatchBusWire(arc,
                                 "BOUNCE",
                                 sizeof("BOUNCE")/sizeof(char) - 1,
                                 3);
      break;
   }
   case Virtex4Cost::PINFEED:
   {
      // IMUX_B0, IMUX_B1, IMUX_B2, ..., IMUX_B31 have cost 2.
      // CLK_B0, CLK_B1, CLK_B2, and CLK_B3 have cost 2.
      onCount = 2
         * (MatchBusWire(arc,
                         "IMUX_B",
                         sizeof("IMUX_B")/sizeof(char) - 1,
                         31)
            + MatchBusWire(arc,
                           "CLK_B",
                           sizeof("CLK_B")/sizeof(char) - 1,
                           3));
      break;
   }
   default:
      onCount = 0;
   };
   return offCount + onCount;
}
